section \<open>Reflexive-transitive closure\<close>

theory Closure
imports Main
begin

section \<open>Inductive definition\<close>

typ "'a rel"  \<comment> \<open>synonym for "('a \<times> 'a) set"\<close>

inductive_set
  star :: "'a rel \<Rightarrow> 'a rel"  ("_\<star>" [100] 100)
  for R :: "'a rel"
where
  base: "(x, x) \<in> R\<star>"
| step: "(x, y) \<in> R \<Longrightarrow> (y, z) \<in> R\<star> \<Longrightarrow> (x, z) \<in> R\<star>"

text \<open>
  This means @{term "R\<star>"} is the least relation (set of pairs) that is closed
  under the introduction rules above. The following rules are derived as
  consequences:
\<close>

term star

thm base step
thm star.cases
thm star.induct

print_statement base step
print_statement star.cases
print_statement star.induct


section \<open>Key properties\<close>

theorem star_refl:
  fixes x :: 'a and R :: "'a rel"
  shows "(x, x) \<in> R\<star>"
  by (rule star.base)

theorem star_trans:
  fixes x y w and R :: "'a rel"
  assumes "(x, y) \<in> R\<star>"
  shows "(y, w) \<in> R\<star> \<Longrightarrow> (x, w) \<in> R\<star>"
  using assms
proof induct
  case (base x)
  then show ?case .
next
  case (step x y z)
  from step(3,4) have "(y, w) \<in> R\<star>" .
  with \<open>(x, y) \<in> R\<close> show "(x, w) \<in> R\<star>" by (rule star.step)
qed

text \<open>Shorter proofs (more implicit)\<close>

lemma "(x, y) \<in> R\<star> \<Longrightarrow> (y, z) \<in> R\<star> \<Longrightarrow> (x, z) \<in> R\<star>"
  \<comment> "weakly structured script"
  apply (induct rule: star.induct)
   apply (auto intro: star.step)
  done

lemma "(x, y) \<in> R\<star> \<Longrightarrow> (y, z) \<in> R\<star> \<Longrightarrow> (x, z) \<in> R\<star>"
  \<comment> "minimal structured proof"
  by (induct rule: star.induct) (auto intro: star.step)


section \<open>Exercise\<close>

text \<open>
  \<triangleright> Prove that @{term "R\<star>"} is the reflexive-transitive closure over @{term "R"}
    \<^item> it is reflexive and transitive
    \<^item> it has the properties of a "Closure operator" (see Wikipedia)

  See also @{term Id} and @{term "op O"} in the main Isabelle/HOL library.
\<close>

declare subrelI [Pure.intro]
  \<comment> "standard introduction for binary relations"

theorem
  fixes R :: "'a rel"
  shows "Id \<subseteq> R\<star>"
proof
  fix x y :: 'a
  assume "(x, y) \<in> Id" then have "x = y" by simp
  then show "(x, y) \<in> R\<star>" by (simp add: star_refl)
qed

theorem
  fixes R :: "'a rel"
  shows "R\<star> O R\<star> \<subseteq> R\<star>"
  sorry

theorem star_extensive:
  fixes R :: "'a rel"
  shows "R \<subseteq> R\<star>"
  sorry

theorem star_increasing:
  fixes R S :: "'a rel"
  assumes "R \<subseteq> S"
  shows "R\<star> \<subseteq> S\<star>"
  sorry

theorem star_idempotent:
  fixes R :: "'a rel"
  shows "R\<star>\<star> = R\<star>"
  sorry


section \<open>Abstract closure operators\<close>

text \<open>
  See also @{url "http://en.wikipedia.org/wiki/Closure_operator"}
  section "Closure operators on partially ordered sets".

  We formalize the main aspect of the prose text as follows:
\<close>

class closure = order +
  fixes cl :: "'a \<Rightarrow> 'a"
  assumes extensive: "x \<le> cl x"
    and increasing: "x \<le> y \<Longrightarrow> cl x \<le> cl y"
    and idempotent: "cl (cl x) = cl x"


section \<open>Exercise\<close>

text \<open>
  The Wikipedia article describes an alternative characterization:
  @{prop "\<And>x y. x \<le> cl y \<longleftrightarrow> cl x \<le> cl y"}

  \<triangleright> Prove the second form as consequence of the above class definition.
\<close>

lemma (in closure) "x \<le> cl y \<longleftrightarrow> cl x \<le> cl y"
proof
  assume "x \<le> cl y"
  show "cl x \<le> cl y" sorry
next
  assume "cl x \<le> cl y"
  show "x \<le> cl y" sorry
qed


section \<open>Exercise\<close>

text \<open>
  \<triangleright> Prove that any operator on partial order with the above property
    is a closure in the sense of the above class definition.

  Note: Take care of proper quantification in assumptions vs. conclusions.
\<close>


section \<open>Exercise\<close>

text \<open>
  \<triangleright> Make alternative proofs using the equivalence rule with the simplifier.

  \<triangleright> What are the technical side-conditions to use the term-rewriting engine
    effectively?
\<close>

end
